(**********************************************************************

:Program.    SortLib.mod
:Contens.    Ein generisches Modul mit schnellen Sortieralgorithmen.
:Author.     Bernd Braun
:Address.    Lippestr. 11, D-3300 Braunschweig
:Phone.      0531/845498
:Copyright.  Public Domain
:Language.   Modula-2
:Translator. M2Amiga A+L V3.32d
:Support.    J. Williams, C. Hoare. 
:History.    V1.0 25.Nov.1990
:History.    V1.1  9.Dez.1990 Kleine Verbesserungen.

***********************************************************************)

IMPLEMENTATION MODULE SortLib;

(* $R- $S- $V- $F- *)

   FROM Arts IMPORT
      Assert;
   FROM Storage IMPORT
      ALLOCATE, DEALLOCATE, Available;
   FROM SYSTEM IMPORT
      ADDRESS, ADR, BYTE, SHIFT;

   CONST
      memmsg   = 'SortLib: Kein Speicher mehr!';

   TYPE
      item     = ARRAY [ 0 .. MAX ( INTEGER ) ] OF BYTE;

   (*
      HeapSort.
      Schneller Sortieralgorithmus von J. Williams 1964.
      Zeitkomplexitt: O ( n log n ) in jedem Fall garantiert.
   *)
   PROCEDURE HeapSort ( VAR a         : ARRAY OF BYTE;
                            comp      : CompareProc;
                            copy      : CopyProc;
                            anz, size : LONGINT );
      VAR
         L, R : LONGINT;
         x    : POINTER TO item;

      (* Linkstes Element in den Heap heruntersinken lassen. *)
      PROCEDURE sift ( L, R : LONGINT );
         VAR
            i, j : LONGINT;
      BEGIN
         i := L;
         j := SHIFT ( i, 1 );
         copy ( x, ADR ( a [ L * size ] ) );
         IF ( j < R ) AND
            ( comp ( ADR ( a [ j * size ] ),
                     ADR ( a [ ( j + 1 ) * size ] ) ) < 0 ) THEN
            INC ( j );
         END;
         WHILE ( j <= R ) AND
               ( comp ( x, ADR ( a [ j * size ] ) ) < 0 ) DO
            copy ( ADR ( a [ i * size ] ), ADR ( a [ j * size ] ) );
            i := j;
            j := SHIFT ( j, 1 );
            IF ( j < R ) AND
               ( comp ( ADR ( a [ j * size ] ),
                        ADR ( a [ ( j + 1 ) * size ] ) ) < 0 ) THEN
               INC ( j );
            END;
         END;
         copy ( ADR ( a [ i * size ] ), x );
      END sift;

      PROCEDURE swap ( Elem1, Elem2 : ADDRESS );
      BEGIN
         copy ( x, Elem1     );
         copy ( Elem1, Elem2 );
         copy ( Elem2, x     );
      END swap;

   BEGIN
      (* Zwischenspeicherelement erzeugen. *)
      Assert ( Available ( size ), ADR ( memmsg ) );
      ALLOCATE ( x, size );
      Assert ( x # NIL, ADR ( memmsg ) );
      L := SHIFT ( anz, - 1 );
      R := anz - 1;
      (* Unregelmig verteilte Elemente in Feld a als Heap organisieren. *)
      WHILE L > 0 DO
         DEC ( L );
         sift ( L, R );
      END;
      (* Feld a sortieren. *)
      WHILE R > 0 DO
         (* Elemente a [ 0 ] und a [ R ] vertauschen. a [ 0 ] ist das
            kleinste Element im Heap. *)
         swap ( ADR ( a [ 0 ] ), ADR ( a [ R * size ] ) );
         DEC ( R );
         sift ( L, R );
      END;
      (* Zwischenspeicherelement lschen. *)
      DEALLOCATE ( x, size );
   END HeapSort;

   (*
      QuickSort.
      Schneller Sortieralgorithmus von C. Hoare 1962.
      Zeitkomplexitt: O ( n^2 )     im schlechtesten Fall.
                       O ( n log n ) im normalen Fall.
      Weil der schlechteste Fall fast nie vorkommt ist QuickSort
      wegen der Einfachheit des Algorithmus ca. 2 bis 3 mal schneller
      als HeapSort.
   *)
   PROCEDURE QuickSort ( VAR a         : ARRAY OF BYTE;
                             comp      : CompareProc;
                             copy      : CopyProc;
                             anz, size : LONGINT );
      VAR
         x : POINTER TO item;

      PROCEDURE swap ( Elem1, Elem2 : ADDRESS );
      BEGIN
         copy ( x, Elem1     );
         copy ( Elem1, Elem2 );
         copy ( Elem2, x     );
      END swap;

      PROCEDURE sort ( L, R : LONGINT );
         VAR
            i, j, midsize, isize, jsize : LONGINT;
      BEGIN
         i := L;
         j := R;
         midsize := size * SHIFT ( i + j, - 1 );
         (* Feld a in zwei Teile zerlegen. *)
         REPEAT
            WHILE comp ( ADR ( a [ i * size ] ), 
                         ADR ( a [ midsize ] ) ) < 0 DO
               INC ( i );
            END;
            WHILE comp ( ADR ( a [ midsize ] ), 
                         ADR ( a [ j * size ] ) ) < 0 DO
               DEC ( j );
            END;
            IF i <= j THEN
               isize := i * size;
               jsize := j * size;
               swap ( ADR ( a [ isize ] ), ADR ( a [ jsize ] ) );
               IF midsize = isize THEN
                  midsize := jsize;
               ELSIF midsize = jsize THEN
                  midsize := isize;
               END;
               INC ( i );
               DEC ( j );
            END;
         UNTIL i > j;
         (* Linken Teil von Feld a sortieren. *)
         IF L < j THEN
            sort ( L, j );
         END;
         (* Rechten Teil von Feld a sortieren. *)
         IF ( i < R ) THEN
            sort ( i, R );
         END;
      END sort;

   BEGIN
      (* Zwischenspeicherelement erzeugen. *)
      Assert ( Available ( size ), ADR ( memmsg ) );
      ALLOCATE ( x, size );
      Assert ( x # NIL, ADR ( memmsg ) );
      (* Sortierung starten. *)
      sort ( 0, anz - 1 );
      (* Zwischenspeicherelement lschen. *)
      DEALLOCATE ( x, size );
   END QuickSort;

   (*
      Gebrauch der Prozeduren:
      a mu ein ARRAY [ 0 .. anz - 1 ] OF Ingentwas sein.
      comp vergleicht zwei Irgentwas.
      copy copiert ein Irgentwas in ein anderes.
      anz = Anzahl der Array-Elemente.
      size = TSIZE ( Irgentwas ).

      Ausfhrliche Beispiele in den Programmen SortLibTest
      und SimpleSort. Dort sind auch comp- und copy-Prozeduren 
      fr INTEGER, Strings und dynamische Strings.
   *)

END SortLib.
