(*
  :Program.       MicroTimer.mod
  :Author.        Volker Rudolph
  :Address.       Lettow-Vorbeck-Str. 11 / 6750 Kaiserslautern 26
  :Phone.         06301/8566
  :Version.       1.4
  :Date.          2.1.1990
  :Copyright.     PD
  :Language.      Oberon
  :Translator.    Amiga-Oberon V1.17.1
  :History.       V1.1 V. Rudolph 24.Apr.1989
  :History.       V1.2 [bne] 03.Sep.1989 (+ CiaPortName)
  :History.       V1.3 V. Rudolph 2.1.1990 (Umsetzung in Oberon, neue Funktionen)
  :History.       V1.4 V. Rudolph 7.3.1990 (Anpassung an Oberon-Compiler V0.9)
  :Contents.      Das Modul MicroTimer ermglicht sehr przise Zeitmessungen.
*)

(* $StackChk- $OvflChk- $RangeChk- $NilChk- $ReturnChk- $CaseChk- $TypeChk- *)
MODULE MicroTimer;

IMPORT h:Hardware,ma:MathIEEEDoubBas,d:Dos,e:Exec,s:SYSTEM,es:ExecSupport;

(* --- EXPORTED ------------------------------------------------------------- *)

CONST
  CiaPortName * ="ciab.timer";

(* PROCEDURE Start*; *)
(* PROCEDURE Look*(VAR ticks:LONGINT); *)
(* PROCEDURE Stop*(VAR ticks:LONGINT); *)
(* PROCEDURE Continue*; *)
(* PROCEDURE TicksToTime*(VAR minutes, seconds:INTEGER;VAR micros:LONGINT; *)

(* --- NOT EXPORTED --------------------------------------------------------- *)

CONST
  INUSE  = "\[33mMicroTimer: Timer of CIAB is already in use\[31m\n";
  NOPORT = "\[33mMicroTimer: Can't create message-port\[31m\n";
  MAXINT = MAX(LONGINT);

  (* Umrechung CIA-Ticks in Mikrosekunden (Systemtakt / 10 in MHz) *)
  TickToMicro = 0.709379;

  (* Maximale Anzahl von Ticks bis zum LONGINT-Overflow *)
  MaxTicks = SHORT(MAXINT * TickToMicro);

TYPE
  (* Konversion: Vier einzelne Bytes in LONGINT *)
  Convert = STRUCT
              byte3:BYTE;
              byte2:BYTE;
              byte1:BYTE;
              byte0:BYTE;
            END;

VAR
  ciaPort:e.MsgPortPtr;
  running:BOOLEAN;
  result:LONGINT;

(* -------------------------------------------------------------------------- *)

(* Timer stoppen *)
PROCEDURE StopIt;
BEGIN
  IF running THEN
    running := FALSE;
    h.ciab.cra := SHORTSET{};
    h.ciab.crb := SHORTSET{};
  END; (* IF *)
END StopIt;

(* -------------------------------------------------------------------------- *)

(* Timer starten *)
PROCEDURE Start*;
BEGIN
  running := TRUE;
  (* Beide Timer auf MAXINT stellen *)
  h.ciab.talo := 0FFX;
  h.ciab.tahi := 0FFX;
  h.ciab.tblo := 0FFX;
  h.ciab.tbhi := 07FX;
  (* Timer b mit Timer a koppeln und Startzeit von Timer b laden *)
  h.ciab.crb := SHORTSET{h.crbLoad,h.crbStart,h.crbRunmode,h.crbInmode1};
  (* Timer a mit Startzeit laden und starten *)
  h.ciab.cra := SHORTSET{h.craLoad,h.craStart};
END Start;

(* -------------------------------------------------------------------------- *)

(* Zeit auslesen whrend Timer luft *)
PROCEDURE Look*(VAR ticks:LONGINT);
VAR
  lPtr:POINTER TO LONGINT;
  conv:Convert;
  ok:BOOLEAN;
BEGIN
  (* Zeit auslesen, bei Fehler wiederholen bis ok *)
  REPEAT
    conv.byte3 := h.ciab.tbhi;
    conv.byte2 := h.ciab.tblo;
    ok :=        (h.ciab.tbhi = conv.byte3);
    conv.byte1 := h.ciab.tahi;
    ok := ok AND (h.ciab.tblo = conv.byte2);
    conv.byte0 := h.ciab.talo;
    ok := ok AND (h.ciab.tahi = conv.byte1);
  UNTIL ok;
  lPtr := s.ADR(conv);

  (* Verstrichene Zeit berechnen *)
  ticks := (MAXINT - lPtr^);

  (* Ist maximale Medauer berschritten ? *)
  IF ticks > MaxTicks THEN
    ticks := 0;
  END; (* IF *)

END Look;

(* -------------------------------------------------------------------------- *)

(* Timer stoppen und auslesen *)
PROCEDURE Stop*(VAR ticks:LONGINT);
VAR
  conv:Convert;
  lPtr:POINTER TO LONGINT;
BEGIN
  StopIt;

  conv.byte3 := h.ciab.tbhi;
  conv.byte2 := h.ciab.tblo;
  conv.byte1 := h.ciab.tahi;
  conv.byte0 := h.ciab.talo;
  lPtr := s.ADR(conv);

  (* Verstrichene Zeit berechnen *)
  ticks := (MAXINT - lPtr^);

  (* Ist maximale Medauer berschritten ? *)
  IF ticks > MaxTicks THEN
    ticks := 0;
  END; (* IF *)
END Stop;

(* -------------------------------------------------------------------------- *)

(* Timer nach Stop() weiterlaufen lassen *)
PROCEDURE Continue*;
BEGIN
  running := TRUE;
  h.ciab.crb := SHORTSET{h.crbStart,h.crbRunmode,h.crbInmode1};
  h.ciab.cra := SHORTSET{h.craStart};
END Continue;

(* -------------------------------------------------------------------------- *)

(* Mikrosekunden in Minuten, Sekunden und restliche Micros umrechnen *)
PROCEDURE TicksToTime*(VAR minutes, seconds:INTEGER;VAR micros:LONGINT;
  ticks:LONGINT);
BEGIN
   ticks   :=  ma.Fix(ma.Flt(ticks) / TickToMicro);
   minutes :=  SHORT(ticks DIV (1000000 * 60));
   seconds :=  SHORT((ticks DIV 1000000) MOD 60);
   micros  :=  ticks MOD 1000000;
END TicksToTime;

(* -------------------------------------------------------------------------- *)

BEGIN
  running := FALSE;
  ciaPort := NIL;

  IF (e.FindPort(CiaPortName) # NIL) OR
     (h.crbStart IN h.ciab.crb) OR
     (h.craStart IN h.ciab.cra) THEN
    IF d.Output() # NIL THEN
      result := d.Write(d.Output(),INUSE,s.SIZE(INUSE));
    END; (* IF *)
    HALT(20);
  END; (* IF *)

  ciaPort := es.CreatePort(s.ADR(CiaPortName),0);
  IF ciaPort = NIL THEN
    IF d.Output() # NIL THEN
      result := d.Write(d.Output(),NOPORT,s.SIZE(NOPORT));
    END; (* IF *)
    HALT(20);
  END; (* IF *)

CLOSE
  StopIt;
  IF ciaPort # NIL THEN
    es.DeletePort(ciaPort);
  END; (* IF *)
END MicroTimer.
