(*
  :Program.       Expressions.mod (OProf)
  :Author.        Volker Rudolph
  :Address.       Lettow-Vorbeck-Str. 11 / 6750 Kaiserslautern 26
  :Phone.         06301/8566
  :Version.       1.22
  :Date.          4.11.90
  :Copyright.     Volker Rudolph (Shareware)
  :Language.      Oberon
  :Translator.    Oberon V1.17.1
  :Imports.       MicroTimer, Printf
  :Contents.      Laufzeit-Statistiken ber Programme
*)

MODULE Expressions;

IMPORT as:ASCII,ob:OberonLib,d:Dos,e:Exec,sd:SecureDos,st:Strings,
  Icons,p:Printf,s:SYSTEM;

(* --- EXPORTED -------------------------------------------------------------- *)

CONST
  MinExpressionLen * = 16;
  MaxExpressionLen * = 1024;
  StringLen * = 80;
  BufSize * = 512;

TYPE
  MaxExpressionPtr * = POINTER TO MaxExpression;
  MinExpression    * = ARRAY MinExpressionLen OF CHAR;
  MaxExpression    * = ARRAY MaxExpressionLen OF CHAR;
  String           * = ARRAY StringLen OF CHAR;

(*
  PROCEDURE IsASCII(ch:CHAR):BOOLEAN;
  PROCEDURE Open*(name:ARRAY OF CHAR;icon:BOOLEAN):BOOLEAN;
  PROCEDURE Close*;
  PROCEDURE ReadExpression(VAR expression:ARRAY OF CHAR;VAR len:INTEGER;VAR semicolon:BOOLEAN);
  PROCEDURE WriteExpression(VAR expression:ARRAY OF CHAR;len:INTEGER):BOOLEAN;
*)

(* --- NOT EXPORTED ---------------------------------------------------------- *)

CONST
  DEBUG = FALSE;

VAR
  buffer:ARRAY BufSize OF CHAR;
  bufIndex:INTEGER;
  bufHigh:INTEGER;
  eof:BOOLEAN;
  actCh:CHAR;
  nextCh:CHAR;

  rHandle:d.FileHandlePtr;
  wHandle:d.FileHandlePtr;

(* -------------------------------------------------------------------------- *)

PROCEDURE IsASCII*(ch:CHAR):BOOLEAN;
BEGIN
  RETURN ((CAP(ch) >= 'A') AND (CAP(ch) <= 'Z')) OR
         ((ch >= '0') AND (ch <= '9')) OR
         (ch = '.');
END IsASCII;

(* -------------------------------------------------------------------------- *)

PROCEDURE Exists(path:ARRAY OF CHAR):BOOLEAN;
VAR
  lock:d.FileLockPtr;
BEGIN
  lock := sd.Lock(path,d.sharedLock);
  IF lock # NIL THEN
    sd.UnLock(lock);
    RETURN TRUE;
  END; (* IF *)
  RETURN FALSE;
END Exists;

PROCEDURE Open*(name:ARRAY OF CHAR;icon:BOOLEAN):BOOLEAN;
VAR
  obPath:String;
  profPath:String;
  path:BOOLEAN;
  i:INTEGER;

BEGIN
  path := FALSE;
  i := st.Length(name)-1;
  WHILE (i >= 0) AND ~path DO
    path := (name[i] = '/') OR (name[i] = ':');
    DEC(i);
  END; (* WHILE *)
  obPath := "";
  profPath := "";
  IF Exists("TXT") AND ~path THEN
    obPath := "TXT/";
    profPath := "TXT/";
  END; (* IF *)
  st.Append(obPath,name);
  st.Append(obPath,".mod");
  st.Append(profPath,name);
  st.Append(profPath,".prf");

  p.Printf1("Creating %s ...\n",s.ADR(profPath));
  rHandle := sd.Open(obPath,d.readOnly);
  IF rHandle = NIL THEN
    RETURN FALSE;
  END; (* IF *)
  wHandle := sd.Open(profPath,d.newFile);
  bufIndex := 0;
  bufHigh := SHORT(d.Read(rHandle,buffer,BufSize));
  eof := wHandle = NIL;
  actCh := as.nul;
  nextCh := as.nul;
  IF icon THEN
    Icons.PutIcon("TXT",profPath);
  END; (* IF *)
  RETURN ~eof;
END Open;

PROCEDURE Close*;
BEGIN
  IF rHandle # NIL THEN
    sd.Close(rHandle);
    rHandle := NIL;
  END; (* IF *)
  IF wHandle # NIL THEN
    sd.Close(wHandle);
    wHandle := NIL;
  END; (* IF *)
END Close;

(* -------------------------------------------------------------------------- *)

PROCEDURE NextChar;
BEGIN
  IF bufIndex = bufHigh THEN
    bufIndex := 0;
    bufHigh := SHORT(d.Read(rHandle,buffer,BufSize));
    eof := bufHigh = 0;
  END; (* IF *)
  actCh := nextCh;
  nextCh := buffer[bufIndex];
  INC(bufIndex);
END NextChar;

PROCEDURE ReadExpression*(VAR expression:ARRAY OF CHAR;VAR len:INTEGER;VAR semicolon:BOOLEAN);
VAR
  i:INTEGER;
  lIndex:INTEGER;
  expressionFull:BOOLEAN;

  PROCEDURE PutChar(ch:CHAR);
  BEGIN
    IF (lIndex < LEN(expression)-2) AND (ch # as.nul) THEN
      IF ch = ';' THEN
        semicolon := TRUE;
      END; (* IF *)
      expression[lIndex] := ch;
      INC(lIndex);
    ELSE
      expressionFull := TRUE;
    END; (* IF *)
  END PutChar;

  PROCEDURE SkipComments;
  BEGIN
    REPEAT
      NextChar;
      PutChar(actCh);
      IF (actCh = '(') AND (nextCh = '*') THEN
        SkipComments;
      END; (* IF *)
    UNTIL ((actCh = '*') AND (nextCh = ')')) OR eof OR expressionFull;
    NextChar;
    PutChar(actCh);
  END SkipComments;

  PROCEDURE SkipString(ch:CHAR);
  BEGIN
    REPEAT
      NextChar;
      PutChar(actCh);
    UNTIL (actCh = ch) OR eof OR expressionFull;
  END SkipString;

  PROCEDURE Next(bracket:CHAR);
  BEGIN
    NextChar;

    IF (actCh = '(') AND (nextCh = '*') THEN
      PutChar(actCh);
      SkipComments;
    ELSE

      IF actCh # as.nul THEN
        PutChar(actCh);
      END; (* IF *)

      CASE actCh OF
        '(':Next(')');
            Next(as.nul);
       |'{':Next('}');
            Next(as.nul);
       |'[':Next(']');
            Next(as.nul);
       |'"':SkipString('"');
       |"'":SkipString("'");
      ELSE
      END; (* CASE *)

      IF bracket # as.nul THEN
        WHILE (actCh # bracket) AND ~eof AND ~expressionFull DO
          Next(as.nul);
        END; (* WHILE *)
      END; (* IF *)
    END; (* IF *)

  END Next;

BEGIN
  semicolon := FALSE;
  len := 0;
  IF bufHigh # 0 THEN
    lIndex := 0;
    expressionFull := FALSE;
    REPEAT
      Next(as.nul);
    UNTIL ~IsASCII(nextCh) OR eof OR expressionFull;
    WHILE ~IsASCII(nextCh) AND ~eof AND ~expressionFull DO
      Next(as.nul);
    END; (* WHILE *)
    expression[lIndex] := as.nul;
    len := lIndex;
  END; (* IF *)
  IF DEBUG THEN
    p.Printf1("ReadExpression <%s>\n",s.ADR(expression));
  END; (* IF *)
END ReadExpression;

PROCEDURE WriteExpression*(VAR expression:ARRAY OF CHAR;len:INTEGER):BOOLEAN;
VAR
  num:INTEGER;
BEGIN
  num := SHORT(d.Write(wHandle,expression,len));
  RETURN num = len;
END WriteExpression;

(* -------------------------------------------------------------------------- *)

BEGIN
  rHandle := NIL;
  wHandle := NIL;
  eof := TRUE;
  bufIndex := 0;
  bufHigh := 0;
END Expressions.
