
(**********************************************************************

    :Program.    TimerSupport
    :Contents.   support to handle the timer.device
    :Author.     Stefan Salewski
    :Address.    Stefan Salewski, Stolper Weg 3, D-2160 Stade
    :Copyright.  Public Domain
    :Language.   Oberon
    :Translator. Amiga-Oberon-Compiler V1.17.1
    :History.    V1.0    01-JAN-1991
    :Support.    Hartmut Hhn & Nicolas Benezan
    :Support.    Procedures SysToNormalTime and NormalToSysTime from AMOK#25

**********************************************************************)

MODULE TimerSupport;

  IMPORT SYSTEM,
         Conversions,
         io,
         Exec,
         ExecSupport,
         Lists;

  CONST
    timerName  * = "timer.device";
    addRequest * = Exec.nonstd+0;
    getSysTime * = Exec.nonstd+1;
    setSysTime * = Exec.nonstd+2;

    (* mode for OpenTimer *)
    microHz * = 0;
    vBlank  * = 1;

    micsPerSec  = 1000000;
    secsPerMin  = 60;
    minsPerHour = 60;
    hoursPerDay = 24;
    daysPerYear = 365;
    daysPerLeap = 4*daysPerYear+1;

  TYPE
    TimeVal * = STRUCT
      secs * : LONGINT;
      micro * : LONGINT
    END;
    TimeRequest * = STRUCT
      node : Exec.IORequest;
      time : TimeVal
    END;
    TimeRequestPtr * = POINTER TO TimeRequest;

    TRNodePtr = POINTER TO TRNode;
    TRNode = RECORD (Lists.Node)
               timeReqPtr:TimeRequestPtr;
             END;

  VAR
    timer:Exec.DevicePtr;
    timerPort:Exec.MsgPortPtr;
    tRList:Lists.List;
    tRPtr:TimeRequestPtr;
    time:TimeVal;

  PROCEDURE AddTime * {timer,-42}(VAR dest{8},source{9}: TimeVal);
  PROCEDURE CmpTime * {timer,-54}(VAR tv1{8} ,tv2{9}   : TimeVal): INTEGER;
  PROCEDURE SubTime * {timer,-48}(VAR dest{8},source{9}: TimeVal);

  PROCEDURE * CloseIt(el:Lists.NodePtr);
  BEGIN
    Exec.CloseDevice(el(TRNode).timeReqPtr);
    ExecSupport.DeleteExtIO(el(TRNode).timeReqPtr);
  END CloseIt;

  PROCEDURE CloseTimer * (timeReqPtr:TimeRequestPtr);
    VAR
      el:Lists.NodePtr;
  BEGIN
    el:=Lists.Head(tRList);
    REPEAT
      IF el(TRNode).timeReqPtr=timeReqPtr THEN
        Lists.Remove(tRList,el);
        Exec.CloseDevice(timeReqPtr);
        ExecSupport.DeleteExtIO(timeReqPtr);
        DISPOSE(el);
        RETURN
      END;
    UNTIL NOT Lists.Next(el);
    HALT(20); (* Timer not open *)
  END CloseTimer;

  PROCEDURE OpenTimer * (mode:LONGINT):TimeRequestPtr;
    VAR
      el:TRNodePtr;
  BEGIN
    NEW(el);
    IF el#NIL THEN
      el.timeReqPtr:=ExecSupport.CreateExtIO(timerPort,
                                             SYSTEM.SIZE(TimeRequest));
      IF el.timeReqPtr#NIL THEN
        IF Exec.OpenDevice(timerName,mode,el.timeReqPtr,LONGSET{})=0 THEN
          Lists.AddHead(tRList,el);
          RETURN el.timeReqPtr
        END;
        ExecSupport.DeleteExtIO(el.timeReqPtr);
      END;
      DISPOSE(el);
    END;
    RETURN NIL;
  END OpenTimer;

  PROCEDURE Delay * (timeReqPtr:TimeRequestPtr;secs,micro:LONGINT);
  BEGIN
    timeReqPtr^.node.command:=addRequest;
    timeReqPtr^.time.secs:=secs;
    timeReqPtr^.time.micro:=micro;
    Exec.DoIO(timeReqPtr);
  END Delay;

  PROCEDURE SendIO * (timeReqPtr:TimeRequestPtr;secs,micro:LONGINT);
  BEGIN
    timeReqPtr^.node.command:=addRequest;
    timeReqPtr^.time.secs:=secs;
    timeReqPtr^.time.micro:=micro;
    Exec.SendIO(timeReqPtr);
  END SendIO;

  PROCEDURE WaitIO * (timeReqPtr:TimeRequestPtr);
  BEGIN
    Exec.WaitIO(timeReqPtr);
  END WaitIO;

  PROCEDURE GetSysTime * (VAR t:TimeVal);
  BEGIN
    tRPtr^.node.command:=getSysTime;
    Exec.DoIO(tRPtr);
    t:=tRPtr^.time
  END GetSysTime;

  PROCEDURE SetSysTime * (t:TimeVal);
  BEGIN
    tRPtr^.node.command:=setSysTime;
    tRPtr^.time:=t;
    Exec.DoIO(tRPtr);
  END SetSysTime;

  PROCEDURE StartClock *;
  BEGIN
    GetSysTime(time);
  END StartClock;

  PROCEDURE ReadClock * (VAR t:TimeVal);
  BEGIN
    GetSysTime(t);
    SubTime(t,time); (* t:=t - time *)
  END ReadClock;

  PROCEDURE WriteTime * (t:TimeVal);
    VAR
      str1:ARRAY 11 OF CHAR;
      str2 :ARRAY 3 OF CHAR;
  BEGIN
    IF NOT Conversions.IntToStr(t.secs,str1,10,10,' ') THEN HALT(0) END;
    io.WriteString(str1);
    io.Write(',');
    IF NOT Conversions.IntToStr(t.micro DIV 10000,str2,10,2,'0') THEN
      HALT(0)
    END;
    io.WriteString(str2);
    io.WriteLn;
  END WriteTime;

  (* The following Procedure is the same as on Amok#25
      1990 by Hartmut Hhn & Nicolas Benezan *)

  PROCEDURE SysToNormalTime * (    SysTime: TimeVal;
                               VAR Year, Month, Day: LONGINT;
                               VAR Hour, Minute, Second: LONGINT);
  BEGIN
    Second:=SysTime.secs;
    Minute:=Second DIV secsPerMin;
    Second:=Second MOD secsPerMin;
    Hour:=Minute DIV minsPerHour;
    Minute:=Minute MOD minsPerHour;
    Day:=Hour DIV hoursPerDay;
    Hour:=Hour MOD hoursPerDay;
    DEC(Day, 790);                    (* offset 1.Jan.78 to 1.Mar.1980 *)
    Year:=(4*Day+3) DIV daysPerLeap;
    DEC(Day, daysPerLeap*Year DIV 4);
    Month:=(5*Day+2) DIV 153;         (* 0=Mar ... 11=Feb next year *)
    DEC(Day,(153*Month+2) DIV 5);
    INC(Day);
    INC(Month, 3);                    (* Mar=3 ... next Feb=14 *)
    IF Month>12 THEN
      INC(Year);
      DEC(Month, 12);
    END;
    INC(Year, 1980);
  END SysToNormalTime;

  (* The following Procedure is the same as on Amok#25
      1990 by Hartmut Hhn & Nicolas Benezan *)

  PROCEDURE NormalToSysTime * (    Year, Month, Day: LONGINT;
                                   Hour, Minute, Second: LONGINT;
                               VAR SysTime: TimeVal);
  BEGIN
    DEC(Year, 1980);
    DEC(Month, 3);                    (* 0=Mar ... 11=Feb next year *)
    IF Month<0 THEN
      INC(Month, 12);
      DEC(Year);
    END;
    DEC(Day);
    INC(Day, (153*Month+2) DIV 5);
    INC(Day, daysPerLeap*Year DIV 4);
    INC(Day, 790);
    INC(Hour, Day*hoursPerDay);
    INC(Minute, Hour*minsPerHour);
    INC(Second, Minute*secsPerMin);
    SysTime.secs:=Second;
    SysTime.micro:=0;
  END NormalToSysTime;

BEGIN
  timerPort:=ExecSupport.CreatePort(SYSTEM.ADR('TimerSupport-Port'),0);
  IF timerPort=NIL THEN HALT(20) END;
  Lists.Init(tRList);
  tRPtr:=OpenTimer(vBlank);
  IF tRPtr=NIL THEN HALT(20) END;
  timer:=tRPtr^.node.device;
CLOSE
  Lists.DoForward(tRList,CloseIt);
  IF timerPort#NIL THEN
    ExecSupport.DeletePort(timerPort)
  END;
END TimerSupport.





